/***************************************************/
/* S3m/Mod player by Daniel Marks                  */
/* (C) 1994 By Daniel Marks and David Jeske        */
/* While we retain the copyright to this code,     */
/* this source code is FREE. You may NOT steal     */
/* the copyright of this code from us.             */
/* You may use it in any way you wish, in          */
/* any product you wish. We respectfully ask that  */
/* you email one of us, if possible, if you        */
/* produce something significant with this code,   */
/* or if you have any bug fixes to contribute.     */
/* We also request that you give credit where      */
/* credit is due if you include part of this code  */
/* in a program of your own.                       */
/*                                                 */
/* email: s3mod@uiuc.edu			   */
/*        jeske@uiuc.edu			   */
/*                                                 */
/* See the associated README file for Thanks       */
/***************************************************/


#define ROT_BUF_SIZE 10240
#define DSP_SPEED 22000


extern int audio;
