



/* s3m.h */



extern int                     audio_buffer_size;
extern uint8                  *audio_start_buffer;
extern uint8                  *audio_end_buffer;
extern uint8                  *audio_curptr;
